{-------------------------------------------------------------------}
{ DSPRIORQ.INC                                                      }
{ Copyright (c) 1993, 1994  Julian M. Bucknall. All Rights Reserved }
{-------------------------------------------------------------------}

{=TPriorityQueue======================================================
A priority queue

Much like an ordinary queue, expect that the smallest data object in
the queue is popped first. Another name for a priority queue is a
heap.

If the Compare method returns values in the 'normal' sense (ie -1 if
Data1 < Data2, etc), then data objects will be popped off smallest
first, ie in increasing order. If Compare returns values in the
'reverse' sense (ie -1 if Data1 > Data2, etc), then elements will be
popped off largest first, ie in decreasing order. Thus by carefully
selecting Compare, this object will provide a min-heap and a max-heap.

19Sep93 JMB
======================================================================}
constructor TPriorityQueue.Init(DataOwner : boolean);
  begin
    if not inherited Init(16, DataOwner) then Fail;

    {This allocation always succeeds}
    Rt := acNewNode(nil);

    Rt^.TLink[CLeft] := Rt;
    Rt^.TLink[CRight] := nil;

    Count := 0;
  end;
{--------}
function  TPriorityQueue.Append(aData : pointer) : boolean;
  var
    Node,
    NewParent : PNode;
  begin
    Append := false;
    Node := acNewNode(aData);
    if Assigned(Node) then
      begin
        if (Count = 1) then
             NewParent := Rt
        else NewParent := pqGetNodeFromIndex(Count shr 1);
        with NewParent^ do
          if Odd(Count) then
               TLink[CRight] := Node
          else TLink[CLeft] := Node;
        Node^.PLink := NewParent;
        pqBubbleUp(Node);
        Append := true;
      end;
  end;
{--------}
function  TPriorityQueue.Clone(DataOwner : boolean;
                               NewCompare : TCompareFunc) : PAbstractContainer;
  var
    NewQueue : PPriorityQueue;
    NodeInx  : longint;
    NewNode  : PNode;
    NewData  : pointer;
    NoDataError : boolean;
  begin
    NewQueue := PPriorityQueue(NewContainer(DataOwner));
    NoDataError := true;
    if Assigned(NewQueue) then
      begin
        NewQueue^.SetCompare(NewCompare);
        if (not IsEmpty) then
          for NodeInx := 1 to Count do
            begin
              if DataOwner then
                NoDataError :=
                   NewQueue^.DupData(pqGetNodeFromIndex(NodeInx)^.Data, NewData)
              else
                NewData := pqGetNodeFromIndex(NodeInx)^.Data;
              if not (NoDataError and NewQueue^.Append(NewData)) then
                begin
                  if NoDataError and DataOwner then
                    NewQueue^.DisposeData(NewData);
                  Dispose(NewQueue, Done);
                  NewQueue := nil;
                  Error(edsCloneError);
                  Break;
                end;
            end;
      end;
    Clone := NewQueue;
  end;
{--------}
procedure TPriorityQueue.Empty;
  begin
    if OwnsData then
      while not IsEmpty do DisposeData(Pop)
    else
      while not IsEmpty do Pop;
    if InDone then
      acDisposeNode(Rt);
  end;
{--------}
function  TPriorityQueue.Examine : pointer;
  begin
    if IsEmpty then
         Examine := nil
    else Examine := Rt^.TLink[CRight]^.Data;
  end;
{--------}
function  TPriorityQueue.NewContainer(DataOwner : boolean) : PAbstractContainer;
  var
    NC : PAbstractContainer;
  begin
    NC := New(PPriorityQueue, Init(DataOwner));
    if Assigned(NC) then
      begin
        NC^.SetCompare(Compare);
        NC^.SetDisposeData(DisposeData);
        NC^.SetDupData(DupData);
      end;
    NewContainer := NC;
  end;
{--------}
function TPriorityQueue.Pop : pointer;
  var
    Root,
    LastNode : PNode;
  begin
    {$IFDEF DEBUG}
    Assert(not IsEmpty, 'TPriorityQueue.Pop');
    {$ENDIF}
    Root := Rt^.TLink[CRight];
    LastNode := pqGetNodeFromIndex(Count);
    Pop := Root^.Data;
    Root^.Data := LastNode^.Data;
    with LastNode^.PLink^ do
      if Odd(Count) then
           TLink[CRight] := nil
      else TLink[CLeft] := nil;
    acDisposeNode(LastNode);
    if not IsEmpty then
      pqTrickleDown(Root);
  end;
{--------}
function TPriorityQueue.Replace(aData : pointer) : pointer;
  begin
    Rt^.Data := aData;
    pqTrickleDown(Rt);
    Replace := Rt^.Data;
    Rt^.Data := nil;
  end;
{--------}
procedure TPriorityQueue.pqBubbleUp(Node : PNode);
  var
    AllDone : boolean;
  begin
    AllDone := false;
    repeat
      with Node^ do
        {If our parent is 'larger' than we are, swap data and move up}
        if (PLink <> Rt) and (Compare(PLink^.Data, Data) > 0) then
          begin
            pqSwapData(PLink, Node);
            Node := PLink;
          end
        else
          AllDone := true;
    until AllDone;
  end;
{--------}
function TPriorityQueue.pqGetNodeFromIndex(Inx : longint) : PNode;
  type
    LH = record L, H : word; end;
  var
    Temp : PNode;
    Mask : longint;
  begin
    {$IFDEF DEBUG}
    Assert((0 < Inx) and (Inx <= Count),
           'TPriorityQueue.pqGetNodeFromIndex');
    {$ENDIF}
    Temp := Rt;

    if (LH(Inx).H = 0) then
      asm
        mov ax, $8000
        mov bx, Inx.Word[0]
        les di, Temp
        jmp @@StartTest

      @@Again:
        shr ax, 1
      @@StartTest:
        test ax, bx
        jz @@Again

      {The first walk is always right}
      @@WalkRight:
        les di, es:[di].TNode.BLink  {BLink is equivalent to TLink[CRight]}

      @@TestForAnotherWalk:
        shr ax, 1
        jz @@AllDone
        test ax, bx
        jnz @@WalkRight

      @@WalkLeft:
        les di, es:[di].TNode.FLink  {FLink is equivalent to TLink[CLeft]}
        jmp @@TestForAnotherWalk

      @@AllDone:
        mov Temp.Word[0], di
        mov Temp.Word[2], es
      end
    else
      begin
        {find first bit in Inx}
        Mask := $40000000;
        while ((Mask and Inx) = 0) do Mask := Mask shr 1;
        {walk the tree:
         if the next bit in Inx is zero go left, otherwise right }
        while (Mask <> 0) do
          begin
            if ((Mask and Inx) = 0) then
                 Temp := Temp^.TLink[CLeft]
            else Temp := Temp^.TLink[CRight];
            Mask := Mask shr 1;
          end;
      end;
    pqGetNodeFromIndex := Temp;
  end;
{--------}
procedure TPriorityQueue.pqSwapData(Node1, Node2 : PNode);
  var
    Temp : pointer;
  begin
    Temp := Node1^.Data;
    Node1^.Data := Node2^.Data;
    Node2^.Data := Temp;
  end;
{--------}
procedure TPriorityQueue.pqTrickleDown(Node : PNode);
  var
    Temp : PNode;
    AllDone : boolean;
  begin
    if not Assigned(Node^.TLink[CLeft]) then Exit;
    AllDone := false;
    repeat
      with Node^ do
        begin
          {Find our 'smaller' child}
          if (not Assigned(TLink[CRight])) or
             (Compare(TLink[CLeft]^.Data, TLink[CRight]^.Data) < 0) then
               Temp := TLink[CLeft]
          else Temp := TLink[CRight];
          {If our 'smaller' child is smaller than we are, swap the data,
           and move down}
          if (Compare(Temp^.Data, Data) < 0) then
            begin
              pqSwapData(Temp, Node);
              Node := Temp;
            end
          else
            AllDone := true;
        end;
    until AllDone or (not Assigned(Node^.TLink[CLeft]));
  end;
{---------------------------------------------------------------------}
