{-------------------------------------------------------------------}
{ DSRBTREE.INC                                                      }
{ Copyright (c) 1993, 1994  Julian M. Bucknall. All Rights Reserved }
{-------------------------------------------------------------------}

type
  LH = record L, H : word; end;

{=Red-black tree helper routines=====================================
These routines help out the red-black tree methods. ColorBlack colors
the cursor black, ColorRed colors the cursor red. IsBlack returns
true if the cursor is black, whereas IsRed returns true if is red.
19Sep93 JMB
======================================================================}
procedure ColorBlack(Cursor : TTreeCursor); near;
  begin
    with GetNode(Cursor)^ do
      LH(PKC).L := LH(PKC).L and $FFFD;
  end;
{--------}
function IsBlack(Cursor : TTreeCursor) : boolean; near;
  begin
    with GetNode(Cursor)^ do
      IsBlack := (LH(PKC).L and 2) = 0;
  end;
{--------}
procedure ColorRed(Cursor : TTreeCursor); near;
  begin
    with GetNode(Cursor)^ do
      LH(PKC).L := LH(PKC).L or 2;
  end;
{--------}
function IsRed(Cursor : TTreeCursor) : boolean; near;
  begin
    with GetNode(Cursor)^ do
      IsRed := (LH(PKC).L and 2) <> 0;
  end;


{=TrbSearchTree======================================================
A red-black binary search tree

A red-black tree is a binary search tree with inbuilt balancing
algorithms during Insert and Delete. This ensures that the tree does
not degenerate into a sorted linked list, maintaining its excellent
search times.

The tree is called red-black because certain data objects are labelled
Black and the others are Red such that (1) every Red data object (that
is not at the root) has a Black parent, (2) each path from leaf to
root has the same number of Black data objects, and (3) each leaf is
Black. This set of rules ensures that the tree is (quite) balanced.

References
  Sedgewick: Algorithms
  Wood: Data Structures, Algorithms, and Performance

PS. I also apologise for the unpolitically correct terminology in this
source code! Thank you, Bryan, for pointing it out, but it's too late
now...

19Sep93 JMB
======================================================================}
function  TrbSearchTree.Delete(Cursor : TTreeCursor) : TTreeCursor;
  var
    Pa, Brother, Nephew1, Nephew2 : TTreeCursor;
    Balanced : boolean;
  begin
    DeletedNodeWasBlack := IsBlack(Cursor);
    Cursor := inherited Delete(Cursor);
    Delete := Cursor;
    repeat
      Balanced := true;
      if DeletedNodeWasBlack then
        if IsRed(Cursor) then
          ColorBlack(Cursor)
        else if not IsRoot(Cursor) then
          begin
            Pa := Parent(Cursor);
            if (Kid(Cursor) = CLeft) then
                 Brother := Right(Pa)
            else Brother := Left(Pa);
            if IsRed(Brother) then
              begin
                if IsBlack(Pa) then
                  ColorBlack(Brother);
                ColorRed(Pa);
                Brother := rbPromote(Brother);
                if (Kid(Cursor) = CLeft) then
                     Cursor := Left(Left(Brother))
                else Cursor := Right(Right(Brother));
                Balanced := false;
              end
            else {Brother is black}
              begin
                if (Kid(Cursor) = CLeft) then
                     Nephew1 := Right(Brother)
                else Nephew1 := Left(Brother);
                if IsRed(Nephew1) then
                  begin
                    ColorBlack(Nephew1);
                    if IsRed(Pa) then
                      ColorRed(Brother);
                    ColorBlack(Pa);
                    Brother := rbPromote(Brother);
                  end
                else {Nephew1 is black}
                  begin
                    if (Kid(Cursor) = CLeft) then
                         Nephew2 := Left(Brother)
                    else Nephew2 := Right(Brother);
                    if IsRed(Nephew2) then
                      begin
                        if IsBlack(Pa) then
                          ColorBlack(Nephew2);
                        ColorBlack(Pa);
                        Nephew2 := rbPromote(rbPromote(Nephew2));
                      end
                    else {Nephew2 is black}
                      if IsRed(Pa) then
                        begin
                          ColorBlack(Pa);
                          ColorRed(Brother);
                        end
                      else {Pa is black}
                        begin
                          ColorRed(Brother);
                          Cursor := Pa;
                          Balanced := false;
                        end;
                  end;
              end;
          end;
    until Balanced;
  end;
{--------}
function  TrbSearchTree.Insert(var Cursor : TTreeCursor; aData : pointer) : boolean;
  var
    Pa, GrandPa, Uncle : TTreeCursor;
    Balanced : boolean;
  begin
    if not inherited Insert(Cursor, aData) then
      Insert := false
    else
      begin
        ColorRed(Cursor);
        repeat
          Balanced := true;
          if not IsRoot(Cursor) then
            begin
              Pa := Parent(Cursor);
              if IsRed(Pa) then
                if IsRoot(Pa) then
                  ColorBlack(Pa)
                else
                  begin
                    GrandPa := Parent(Pa);
                    ColorRed(GrandPa);
                    if (Kid(Pa) = CLeft) then
                         Uncle := Right(GrandPa)
                    else Uncle := Left(GrandPa);
                    if IsRed(Uncle) then
                      begin
                        ColorBlack(Pa);
                        ColorBlack(Uncle);
                        Cursor := GrandPa;
                        Balanced := false;
                      end
                    else {Uncle is black}
                      if (Kid(Cursor) = Kid(Pa)) then
                        begin
                          ColorBlack(Pa);
                          Pa := rbPromote(Pa);
                        end
                      else
                        begin
                          ColorBlack(Cursor);
                          Cursor := rbPromote(rbPromote(Cursor));
                        end;
                  end;
            end;
        until Balanced;
        Insert := true;
      end;
  end;
{--------}
function  TrbSearchTree.NewContainer(DataOwner : boolean) : PAbstractContainer;
  var
    NC : PAbstractContainer;
  begin
    NC := New(PrbSearchTree, Init(DataOwner));
    if Assigned(NC) then
      begin
        NC^.SetCompare(Compare);
        NC^.SetDisposeData(DisposeData);
        NC^.SetDupData(DupData);
      end;
    NewContainer := NC;
  end;
{--------}
function TrbSearchTree.rbPromote(Cursor : TTreeCursor) : TTreeCursor;
  var
    NodeX,
    NodeP,
    XSon  : PNode;
  begin
    NodeX := GetNode(Cursor);
    NodeP := Dad(Cursor);

    with NodeP^ do
      begin
        Dad(PKC)^.TLink[Kid(PKC)] := NodeX;
        NodeX^.PKC := Dye(PKC, NodeX^.PKC);
      end;

    if (Kid(Cursor) = CLeft) then
      begin
        XSon := NodeX^.TLink[CRight];
        NodeX^.TLink[CRight] := NodeP;
        NodeP^.PKC := Dye(Csr(NodeX, CRight), NodeP^.PKC);
        NodeP^.TLink[CLeft] := XSon;
        if (XSon <> NilNode) then
          XSon^.PKC := Dye(Cursor, XSon^.PKC);
      end
    else
      begin
        XSon := NodeX^.TLink[CLeft];
        NodeX^.TLink[CLeft] := NodeP;
        NodeP^.PKC := Dye(Csr(NodeX, CLeft), NodeP^.PKC);
        NodeP^.TLink[CRight] := XSon;
        if (XSon <> NilNode) then
          XSon^.PKC := Dye(Cursor, XSon^.PKC);
      end;

    rbPromote := Bleach(NodeX^.PKC);
  end;
{--------}
procedure TrbSearchTree.bsSwapData(OldCursor, NewCursor : TTreeCursor);
  begin
    DeletedNodeWasBlack :=  IsBlack(NewCursor);
    inherited bsSwapData(OldCursor, NewCursor);
  end;
{---------------------------------------------------------------------}
