{-------------------------------------------------------------------}
{ DSSKPLST.INC                                                      }
{ Copyright (c) 1993, 1994  Julian M. Bucknall. All Rights Reserved }
{-------------------------------------------------------------------}

{-A FirstThat iterator for cloning a skip list}
function FailsSkipCloning(SL : PAbstractContainer; aData : pointer;
                          NSL : pointer) : boolean;
  var
    NewList : PSkipList absolute NSL;
    DummyCursor : TListCursor;
    NewData : pointer;
    NoDataError : boolean;
  begin
    FailsSkipCloning := true;
    NoDataError := true;
    with NewList^ do
      begin
        if OwnsData then
             NoDataError := DupData(aData, NewData)
        else NewData := aData;
        if NoDataError then
          if Insert(DummyCursor, NewData) then
               FailsSkipCloning := false
          else if OwnsData then DisposeData(NewData);
      end;
  end;


{=TSkipList===========================================================
A skip linked list

This is a special type of linked list of data objects. Compared with
TList and TDList, this implementation uses nodes of varying sizes. The
nodes have between 1 and 16 (skMaxLevels) of forward pointers, the
higher ones skipping over nodes with less forward pointers. This means
much faster search times, but slightly slower list update times (ie
insert and delete). Can cope with searching long lists without too
much degradation. Compared with a red-black binary search tree, this
type of data structure will consume more memory, will have faster
insert times, slower (?) delete times, and will have comparable
(amortised) search times.

Reference
  Scheiner: Skip Lists (DDJ January 1994)

04Dec93 JMB
======================================================================}
constructor TSkipList.Init(DataOwner : boolean);
  var
    Level : integer;
  begin
    {Note: we cannot use a NodeStore as the nodes have different
           sizes, so pass 0.}
    if not inherited Init(0, DataOwner) then Fail;
    skNewNodeLevel := skMaxLevels;
    BF := acNewNode(nil);
    skNewNodeLevel := 1;
    AL := acNewNode(nil);
    if (not Assigned(BF)) or (not Assigned(AL)) then
      begin
        ConstructorError := edsOutOfMemory;
        Done;
        Fail;
      end;
    for Level := 0 to pred(skMaxLevels) do
      BF^.FwLink[Level] := AL;
    BF^.BkLink:= BF;
    AL^.FwLink[0] := AL;
    AL^.BkLink:= BF;
    CurLevels := 1;
    Count := 0;
  end;
{--------}
function  TSkipList.Clone(DataOwner : boolean;
                          NewCompare : TCompareFunc) : PAbstractContainer;
  var
    NewList : PSkipList;
  begin
    NewList := PSkipList(NewContainer(DataOwner));
    if Assigned(NewList) then
      begin
        NewList^.SetCompare(NewCompare);
        if (not IsEmpty) then
          begin
            if (FirstThat(FailsSkipCloning, NewList) <> 0) then
              begin
                Dispose(NewList, Done);
                NewList := nil;
                Error(edsCloneError);
              end;
          end;
      end;
    Clone := NewList;
  end;
{--------}
function TSkipList.Delete(Cursor : TListCursor) : TListCursor;
  var
    aData     : pointer;
    Walker    : PNode;
    NextStep  : TListCursor;
    TempNode  : PNode;
    Level     : integer;
    CompResult: integer;
    PrevLink  : array [0..pred(skMaxLevels)] of PNode;
  begin
    {$IFDEF DEBUG}
    Assert((not IsBeforeFirst(Cursor)) and (not IsAfterLast(Cursor)),
           'TSkipList.Delete');
    {$ENDIF}
    aData := Examine(Cursor);
    Walker := PNode(SetBeforeFirst);
    for Level := pred(CurLevels) downto 0 do
      begin
        NextStep := TListCursor(Walker^.FwLink[Level]);
        if IsAfterLast(NextStep) then
             CompResult := -1
        else CompResult := Compare(aData, Examine(NextStep));
        while (CompResult > 0) do
          begin
            Walker := PNode(NextStep);
            NextStep := TListCursor(Walker^.FwLink[Level]);
            if IsAfterLast(NextStep) then
                 CompResult := -1
            else CompResult := Compare(aData, Examine(NextStep));
          end;
        PrevLink[Level] := Walker;
      end;
    with PNode(Cursor)^ do
      begin
        TempNode := FwLink[0];
        TempNode^.BkLink := BkLink;
        PrevLink[0]^.FwLink[0] := FwLink[0];
        for Level := 1 to pred(Lvls) do
          PrevLink[Level]^.FwLink[Level] := FwLink[Level];
      end;
    acDisposeNode(PNode(Cursor));
    Delete := TListCursor(TempNode);
  end;
{--------}
procedure TSkipList.Empty;
  var
    Temp,
    Cursor : TListCursor;
    Level  : integer;
  begin
    {Note: it will be faster to delete nodes from first principles
           rather than call the Erase method.}
    Cursor := Next(SetBeforeFirst);
    while not IsAfterLast(Cursor) do
      begin
        Temp := Cursor;
        Cursor := Next(Cursor);
        if OwnsData then
          DisposeData(Examine(Temp));
        acDisposeNode(PNode(Temp));
      end;
    if InDone then
      begin
        if Assigned(BF) then
          acDisposeNode(BF);
        if Assigned(AL) then
          acDisposeNode(AL);
      end
    else
      begin
        {patch everything up again}
        for Level := 0 to pred(skMaxLevels) do
          BF^.FwLink[Level] := AL;
        AL^.BkLink:= BF;
        CurLevels := 1;
        Count := 0;
      end;
  end;
{--------}
function TSkipList.Erase(Cursor : TListCursor) : TListCursor;
  var
    Data : pointer;
  begin
    {Note: Delete requires the Data field so dispose the data
           afterwards}
    Data := Examine(Cursor);
    Erase := Delete(Cursor);
    if OwnsData then
      DisposeData(Data);
  end;
{--------}
function  TSkipList.Examine(Cursor : TListCursor) : pointer;
  begin
    Examine := PNode(Cursor)^.Data;
  end;
{--------}
function  TSkipList.FirstThat(Action : TIteratorFunc; ExtraData : pointer) : TListCursor;
  var
    FoundIt : boolean;
    Cursor  : TListCursor;
  begin
    FoundIt := false;
    Cursor := Next(SetBeforeFirst);
    while (not IsAfterLast(Cursor)) and (not FoundIt) do
      if Action(@Self, Examine(Cursor), ExtraData) then
           FoundIt := true
      else Cursor := Next(Cursor);
    if FoundIt then
         FirstThat := Cursor
    else FirstThat := 0;
  end;
{--------}
procedure TSkipList.ForEach(Action : TIteratorProc; ExtraData : pointer);
  var
    Cursor : TListCursor;
  begin
    Cursor := Next(SetBeforeFirst);
    while not IsAfterLast(Cursor) do
      begin
        Action(@Self, Examine(Cursor), ExtraData);
        Cursor := Next(Cursor);
      end;
  end;
{--------}
function TSkipList.Insert(var Cursor : TListCursor; aData : pointer) : boolean;
  var
    Walker    : PNode;
    NextStep  : TListCursor;
    TempNode  : PNode;
    Level     : integer;
    CompResult: integer;
    PrevLink  : array [0..pred(skMaxLevels)] of PNode;
  begin
    Insert := false;
    Walker := PNode(SetBeforeFirst);
    {note: the following for loop is executed at least once
           because CurLevels >= 1}
    for Level := pred(CurLevels) downto 0 do
      begin
        NextStep := TListCursor(Walker^.FwLink[Level]);
        if IsAfterLast(NextStep) then
             CompResult := -1
        else CompResult := Compare(aData, Examine(NextStep));
        while (CompResult > 0) do
          begin
            Walker := PNode(NextStep);
            NextStep := TListCursor(Walker^.FwLink[Level]);
            if IsAfterLast(NextStep) then
                 CompResult := -1
            else CompResult := Compare(aData, Examine(NextStep));
          end;
        PrevLink[Level] := Walker;
      end;
    if (CompResult = 0) then
      begin
        Error(edsInsertDup);
        Exit;
      end;
    Level := 1;
    while (Level < skMaxLevels) and (jRandom(4) = 0) do
      inc(Level);
    if (Level > CurLevels) then
      begin
        PrevLink[CurLevels] := BF;
        inc(CurLevels);
        Level := CurLevels;
      end;
    skNewNodeLevel := Level;
    TempNode := acNewNode(aData);
    if Assigned(TempNode) then
      begin
        for Level := pred(skNewNodeLevel) downto 0 do
          with PrevLink[Level]^ do
            begin
              TempNode^.FwLink[Level] := FwLink[Level];
              FwLink[Level] := TempNode;
            end;
        with TempNode^.FwLink[0]^ do
          begin
            TempNode^.BkLink := BkLink;
            BkLink := TempNode;
          end;
        Cursor := TListCursor(TempNode);
        Insert := true;
      end;
  end;
{--------}
function  TSkipList.IsAfterLast(Cursor : TListCursor) : boolean;
  begin
    IsAfterLast := (PNode(Cursor) = AL);
  end;
{--------}
function  TSkipList.IsBeforeFirst(Cursor : TListCursor) : boolean;
  begin
    IsBeforeFirst := (PNode(Cursor) = BF);
  end;
{--------}
procedure TSkipList.Join(List : PSkipList);
  var
    Walker,
    Dummy  : TListCursor;
    Data   : pointer;
  begin
    if Assigned(List) then
      with List^ do
        begin
          Walker := Next(SetBeforeFirst);
          while not IsEmpty do
            begin
              Data := Examine(Walker);
              Walker := Delete(Walker);
              if not Self.Insert(Dummy, Data) then
                {nothing - it cannot happen};
            end;
          Dispose(List, Done);
        end;
  end;
{--------}
function  TSkipList.NewContainer(DataOwner : boolean) : PAbstractContainer;
  var
    NC : PAbstractContainer;
  begin
    NC := New(PSkipList, Init(DataOwner));
    if Assigned(NC) then
      begin
        NC^.SetCompare(Compare);
        NC^.SetDisposeData(DisposeData);
        NC^.SetDupData(DupData);
      end;
    NewContainer := NC;
  end;
{--------}
function  TSkipList.Next(Cursor : TListCursor) : TListCursor;
  begin
    Next := TListCursor(PNode(Cursor)^.FwLink[0]);
  end;
{--------}
function  TSkipList.Prev(Cursor : TListCursor) : TListCursor;
  begin
    Prev := TLIstCursor(PNode(Cursor)^.BkLink);
  end;
{--------}
function  TSkipList.Replace(Cursor : TListCursor; aData : pointer) : pointer;
  begin
    if not (IsBeforeFirst(Cursor) or IsAfterLast(Cursor)) then
      begin
        Replace := Examine(Cursor);
        Cursor := Delete(Cursor);
        Insert(Cursor, aData);
      end;
  end;
{--------}
function  TSkipList.Search(var Cursor : TListCursor; aData : pointer) : boolean;
  var
    Walker    : PNode;
    NextStep  : TListCursor;
    Level     : integer;
    CompResult: integer;
  begin
    Walker := PNode(SetBeforeFirst);
    for Level := pred(CurLevels) downto 0 do
      begin
        NextStep := TListCursor(Walker^.FwLink[Level]);
        if IsAfterLast(NextStep) then
             CompResult := -1
        else CompResult := Compare(aData, Examine(NextStep));
        while (CompResult > 0) do
          begin
            Walker := PNode(NextStep);
            NextStep := TListCursor(Walker^.FwLink[Level]);
            if IsAfterLast(NextStep) then
                 CompResult := -1
            else CompResult := Compare(aData, Examine(NextStep));
          end;
      end;
    Cursor := NextStep;
    Search := (CompResult = 0);
  end;
{--------}
function  TSkipList.SetBeforeFirst : TListCursor;
  begin
    SetBeforeFirst := TListCursor(BF);
  end;
{--------}
function  TSkipList.SetAfterLast : TListCursor;
  begin
    SetAfterLast := TListCursor(AL);
  end;
{--------}
function  TSkipList.Split(Cursor : TListCursor) : PSkipList;
  var
    NewList   : PSkipList;
    Dummy,
    NextCursor: TListCursor;
    Data      : pointer;
  begin
    {$IFDEF DEBUG}
    Assert(not IsAfterLast(Cursor), 'TSkipList.Split');
    {$ENDIF}
    NewList := PSkipList(NewContainer(OwnsData));
    if not Assigned(NewList) then
      Error(edsOutOfMemory)
    else
      begin
        NextCursor := Next(Cursor);
        while not IsAfterLast(NextCursor) do
          begin
            Data := Examine(NextCursor);
            NextCursor := Delete(NextCursor);
            if not NewList^.Insert(Dummy, Data) then
              {nothing - it cannot happen};
          end;
      end;
    Split := NewList;
  end;
{--------}
procedure TSkipList.acDisposeNode(aNode : PNode);
  begin
    {$IFDEF DEBUG}
    Assert(Assigned(aNode), 'TSkipList.acDisposeNode');
    {$ENDIF}
    SafeFreeMem(aNode, aNode^.Size);
    dec(Count);
  end;
{--------}
function  TSkipList.acNewNode(aData : pointer) : PNode;
  var
    Node : PNode;
    NodeBytes : word;
  begin
    {Note: we must override the default node allocation as the nodes
           vary in size. The object variable NewNodeLevel is the
           number of forward links we must reserve.}
    {$IFDEF DEBUG}
    Assert((0 < skNewNodeLevel) and (skNewNodeLevel <= skMaxLevels),
           'TSkipList.acNewNode');
    {$ENDIF}
    NodeBytes := (skNewNodeLevel+2) * sizeof(pointer) +
                 (sizeof(word) * 2);
    if not SafeGetMem(Node, NodeBytes) then
      Error(edsOutOfMemory)
    else
      begin
        with Node^ do
          begin
            Data := aData;
            Size := NodeBytes;
            Lvls := skNewNodeLevel;
          end;
        inc(Count);
        {$IFDEF MaxIs64KItems}
        if (Count = 0) then {overflowed word}
          begin
            acDisposeNode(Node);
            Error(edsTooManyItems);
            Node := nil;
          end;
        {$ENDIF}
      end;
    acNewNode := Node;
  end;
{---------------------------------------------------------------------}

